@@ -18,13 +18,13 @@
 #include <malloc.h>
 #include <stdio.h>
 #include <string.h>
-#include <unistd.h>
 #include <signal.h>
 #include <sys/param.h>
 #include <sys/stat.h>
 #include <sys/times.h>
 #include <sys/types.h>
 #include <sys/wait.h>
+#include <poll.h>
 
 #define MODULE_LOOKUP
 #include "automount.h"
@@ -51,28 +51,26 @@ int lookup_init(const char *mapfmt, int argc, const char *const *argv, void **co
 	ctxt = malloc(sizeof(struct lookup_context));
 	if (!ctxt) {
 		char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
-		crit(LOGOPT_ANY, MODPREFIX "malloc: %s", estr);
+		logerr(MODPREFIX "malloc: %s", estr);
 		return 1;
 	}
 
 	if (argc < 1) {
-		crit(LOGOPT_ANY, MODPREFIX "No map name");
+		logmsg(MODPREFIX "No map name");
 		free(ctxt);
 		return 1;
 	}
 	ctxt->mapname = argv[0];
 
 	if (ctxt->mapname[0] != '/') {
-		crit(LOGOPT_ANY,
-		     MODPREFIX "program map %s is not an absolute pathname",
+		logmsg(MODPREFIX "program map %s is not an absolute pathname",
 		     ctxt->mapname);
 		free(ctxt);
 		return 1;
 	}
 
 	if (access(ctxt->mapname, X_OK)) {
-		crit(LOGOPT_ANY,
-		     MODPREFIX "program map %s missing or not executable",
+		logmsg(MODPREFIX "program map %s missing or not executable",
 		     ctxt->mapname);
 		free(ctxt);
 		return 1;
@@ -83,7 +81,7 @@ int lookup_init(const char *mapfmt, int argc, const char *const *argv, void **co
 
 	ctxt->parse = open_parse(mapfmt, MODPREFIX, argc - 1, argv + 1);
 	if (!ctxt->parse) {
-		crit(LOGOPT_ANY, MODPREFIX "failed to open parse context");
+		logmsg(MODPREFIX "failed to open parse context");
 		free(ctxt);
 		return 1;
 	}
@@ -116,14 +114,12 @@ int lookup_mount(struct autofs_point *ap, const char *name, int name_len, void *
 	char errbuf[1024], *errp;
 	char ch;
 	int pipefd[2], epipefd[2];
+	struct pollfd pfd[2];
 	pid_t f;
-	int files_left;
 	int status;
-	fd_set readfds, ourfds;
 	enum state { st_space, st_map, st_done } state;
 	int quoted = 0;
 	int ret = 1;
-	int max_fd;
 	int distance;
 	int alloci = 1;
 
@@ -133,8 +129,23 @@ int lookup_mount(struct autofs_point *ap, const char *name, int name_len, void *
 
 	mc = source->mc;
 
+	/* Check if we recorded a mount fail for this key anywhere */
+	me = lookup_source_mapent(ap, name, LKP_DISTINCT);
+	if (me) {
+		if (me->status >= time(NULL)) {
+			cache_unlock(me->mc);
+			return NSS_STATUS_NOTFOUND;
+		}
+
+		/* Negative timeout expired for non-existent entry. */
+		if (!me->mapent)
+			cache_delete(me->mc, name);
+
+		cache_unlock(me->mc);
+	}
+
 	/* Catch installed direct offset triggers */
-	cache_readlock(mc);
+	cache_writelock(mc);
 	me = cache_lookup_distinct(mc, name);
 	if (!me) {
 		cache_unlock(mc);
@@ -148,22 +159,47 @@ int lookup_mount(struct autofs_point *ap, const char *name, int name_len, void *
 			return NSS_STATUS_NOTFOUND;
 		}
 	} else {
-		cache_unlock(mc);
 		/* Otherwise we found a valid offset so try mount it */
 		debug(ap->logopt, MODPREFIX "%s -> %s", name, me->mapent);
 
-		master_source_current_wait(ap->entry);
-		ap->entry->current = source;
+		/*
+		 * If this is a request for an offset mount (whose entry
+		 * must be present in the cache to be valid) or the entry
+		 * is newer than the negative timeout value then just
+		 * try and mount it. Otherwise try and remove it and
+		 * proceed with the program map lookup.
+		 */
+		if (strchr(name, '/') ||
+		    me->age + ap->negative_timeout > time(NULL)) {
+			char *ent = NULL;
 
-		ret = ctxt->parse->parse_mount(ap, name, name_len,
-				      me->mapent, ctxt->parse->context);
-		goto out_free;
+			if (me->mapent) {
+				ent = alloca(strlen(me->mapent) + 1);
+				strcpy(ent, me->mapent);
+			}
+			cache_unlock(mc);
+			master_source_current_wait(ap->entry);
+			ap->entry->current = source;
+			ret = ctxt->parse->parse_mount(ap, name,
+				 name_len, ent, ctxt->parse->context);
+			goto out_free;
+		} else {
+			if (me->multi) {
+				cache_unlock(mc);
+				warn(ap->logopt, MODPREFIX
+				     "unexpected lookup for active multi-mount"
+				     " key %s, returning fail", name);
+				return NSS_STATUS_UNAVAIL;
+			}
+			cache_delete(mc, name);
+			cache_unlock(mc);
+		}
 	}
 
 	mapent = (char *) malloc(MAPENT_MAX_LEN + 1);
 	if (!mapent) {
 		char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
-		error(ap->logopt, MODPREFIX "malloc: %s", estr);
+		logerr(MODPREFIX "malloc: %s", estr);
 		return NSS_STATUS_UNAVAIL;
 	}
 
@@ -174,12 +210,12 @@ int lookup_mount(struct autofs_point *ap, const char *name, int name_len, void *
 	 * want to send stderr to the syslog, and we don't use spawnl()
 	 * because we need the pipe hooks
 	 */
-	if (pipe(pipefd)) {
+	if (open_pipe(pipefd)) {
 		char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
-		error(ap->logopt, MODPREFIX "pipe: %s", estr);
+		logerr(MODPREFIX "pipe: %s", estr);
 		goto out_free;
 	}
-	if (pipe(epipefd)) {
+	if (open_pipe(epipefd)) {
 		close(pipefd[0]);
 		close(pipefd[1]);
 		goto out_free;
@@ -188,7 +224,7 @@ int lookup_mount(struct autofs_point *ap, const char *name, int name_len, void *
 	f = fork();
 	if (f < 0) {
 		char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
-		error(ap->logopt, MODPREFIX "fork: %s", estr);
+		logerr(MODPREFIX "fork: %s", estr);
 		close(pipefd[0]);
 		close(pipefd[1]);
 		close(epipefd[0]);
@@ -216,30 +252,39 @@ int lookup_mount(struct autofs_point *ap, const char *name, int name_len, void *
 	errp = errbuf;
 	state = st_space;
 
-	FD_ZERO(&ourfds);
-	FD_SET(pipefd[0], &ourfds);
-	FD_SET(epipefd[0], &ourfds);
+	pfd[0].fd = pipefd[0];
+	pfd[0].events = POLLIN;
+	pfd[1].fd = epipefd[0];
+	pfd[1].events = POLLIN;
 
-	max_fd = pipefd[0] > epipefd[0] ? pipefd[0] : epipefd[0];
+	while (1) {
+		int bytes;
 
-	files_left = 2;
+		if (poll(pfd, 2, -1) < 0 && errno != EINTR)
+			break;
 
-	while (files_left != 0) {
-		readfds = ourfds;
-		if (select(max_fd + 1, &readfds, NULL, NULL, NULL) < 0 && errno != EINTR)
+		if (pfd[0].fd == -1 && pfd[1].fd == -1)
+			break;
+
+		if ((pfd[0].revents & (POLLIN|POLLHUP)) == POLLHUP &&
+		    (pfd[1].revents & (POLLIN|POLLHUP)) == POLLHUP)
 			break;
 
 		/* Parse maps from stdout */
-		if (FD_ISSET(pipefd[0], &readfds)) {
-			if (read(pipefd[0], &ch, 1) < 1) {
-				FD_CLR(pipefd[0], &ourfds);
-				files_left--;
+		if (pfd[0].revents) {
+cont:
+			bytes = read(pipefd[0], &ch, 1);
+			if (bytes == 0)
+				goto next;
+			else if (bytes < 0) {
+				pfd[0].fd = -1;
 				state = st_done;
+				goto next;
 			}
 
 			if (!quoted && ch == '\\') {
 				quoted = 1;
-				continue;
+				goto cont;
 			}
 
 			switch (state) {
@@ -271,8 +316,7 @@ int lookup_mount(struct autofs_point *ap, const char *name, int name_len, void *
 						       ++alloci));
 					if (!tmp) {
 						alloci--;
-						error(ap->logopt,
-						      MODPREFIX "realloc: %s",
+						logerr(MODPREFIX "realloc: %s",
 						      strerror(errno));
 						break;
 					}
@@ -297,26 +341,33 @@ int lookup_mount(struct autofs_point *ap, const char *name, int name_len, void *
 				/* Eat characters till there's no more output */
 				break;
 			}
+			quoted = 0;
+			goto cont;
 		}
 		quoted = 0;
-
+next:
 		/* Deal with stderr */
-		if (FD_ISSET(epipefd[0], &readfds)) {
-			if (read(epipefd[0], &ch, 1) < 1) {
-				FD_CLR(epipefd[0], &ourfds);
-				files_left--;
-			} else if (ch == '\n') {
-				*errp = '\0';
-				if (errbuf[0])
-					error(ap->logopt, ">> %s", errbuf);
-				errp = errbuf;
-			} else {
-				if (errp >= &errbuf[1023]) {
+		if (pfd[1].revents) {
+			while (1) {
+				bytes = read(epipefd[0], &ch, 1);
+				if (bytes == 0)
+					break;
+				else if (bytes < 0) {
+					pfd[1].fd = -1;
+					break;
+				} else if (ch == '\n') {
 					*errp = '\0';
-					error(ap->logopt, ">> %s", errbuf);
+					if (errbuf[0])
+						logmsg(">> %s", errbuf);
 					errp = errbuf;
+				} else {
+					if (errp >= &errbuf[1023]) {
+						*errp = '\0';
+						logmsg(">> %s", errbuf);
+						errp = errbuf;
+					}
+					*(errp++) = ch;
 				}
-				*(errp++) = ch;
 			}
 		}
 	}
@@ -325,7 +376,7 @@ int lookup_mount(struct autofs_point *ap, const char *name, int name_len, void *
 		*mapp = '\0';
 	if (errp > errbuf) {
 		*errp = '\0';
-		error(ap->logopt, ">> %s", errbuf);
+		logmsg(">> %s", errbuf);
 	}
 
 	close(pipefd[0]);
@@ -333,20 +384,22 @@ int lookup_mount(struct autofs_point *ap, const char *name, int name_len, void *
 
 	if (waitpid(f, &status, 0) != f) {
 		char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
-		error(ap->logopt, MODPREFIX "waitpid: %s", estr);
+		logerr(MODPREFIX "waitpid: %s", estr);
 		goto out_free;
 	}
 
 	if (mapp == mapent || !WIFEXITED(status) || WEXITSTATUS(status) != 0) {
-		msg(MODPREFIX "lookup for %s failed", name);
+		info(ap->logopt, MODPREFIX "lookup for %s failed", name);
 		goto out_free;
 	}
 
 	cache_writelock(mc);
 	ret = cache_update(mc, source, name, mapent, time(NULL));
 	cache_unlock(mc);
-	if (ret == CHE_FAIL)
+	if (ret == CHE_FAIL) {
+		free(mapent);
 		return NSS_STATUS_UNAVAIL;
+	}
 
 	debug(ap->logopt, MODPREFIX "%s -> %s", name, mapent);
 
@@ -359,8 +412,21 @@ out_free:
 	if (mapent)
 		free(mapent);
 
-	if (ret)
-		return NSS_STATUS_UNAVAIL;
+	if (ret) {
+		time_t now = time(NULL);
+		int rv = CHE_OK;
+
+		cache_writelock(mc);
+		me = cache_lookup_distinct(mc, name);
+		if (!me)
+			rv = cache_update(mc, source, name, NULL, now);
+		if (rv != CHE_FAIL) {
+			me = cache_lookup_distinct(mc, name);
+			me->status = now + ap->negative_timeout;
+		}
+		cache_unlock(mc);
+		return NSS_STATUS_TRYAGAIN;
+	}
 
 	return NSS_STATUS_SUCCESS;
 }
